#!/bin/sh
#
# geninitrd mod: dm-multipath

# if we should init dm-multipath at boot
have_multipath=no

if [ -x /sbin/multipath ]; then
	USE_MULTIPATH=yes
else
	USE_MULTIPATH=no
fi

# dm-multipath wwid which is used for rootfs
MPATH_WWID=

# return true if node is multipath controlled
# @param	string $node device node to be examined
# @access	public
is_multipath() {
	local devpath="$1"

	# multipath disabled
	if ! is_yes "$USE_MULTIPATH"; then
		return 1
	fi

	# multipath nodes are under device mapper
	if [[ "$devpath" != /dev/mapper/* ]]; then
		return 1
	fi

	DM_NAME=
	eval $(dm_export "$devpath")
	if [ -z "$DM_NAME" ]; then
		die "dm_export failed unexpectedly"
	fi

	local MPATH_WWID=${DM_UUID##*-}
	local info=$(multipath -l $MPATH_WWID)
	if [ -z "$info" ]; then
		return 1
	fi

	return 0
}

# find modules for $devpath
# @param	$devpath	device to be examined
# @access	public
# find dm-multipath modules for $devpath
# returns false if $devpath is not dm-multipath
find_modules_multipath() {
	local devpath="$1"

	DM_NAME=
	eval $(dm_export "$devpath")
	if [ -z "$DM_NAME" ]; then
		die "dm_export failed unexpectedly"
	fi

	# Partition:
	#  DM_NAME=LUN-28p1
	#  DM_UUID=part1-mpath-36006016002c11800a0aa05fbfae0db11
	# Disk:
	#  DM_NAME=LUN-28
	#  DM_UUID=mpath-36006016002c11800a0aa05fbfae0db11
	MPATH_WWID=${DM_UUID##*-}

	local info=$(multipath -l $MPATH_WWID)
	if [ -z "$info" ]; then
		return 1
	fi

	debug "Finding modules for dm-multipath (WWID=$MPATH_WWID)"
	have_multipath=yes
	local dev phydevs=$(echo "$info" | awk '$2 ~ /^[0-9]+:[0-9]+:[0-9]+:[0-9]+$/{printf("/dev/%s\n", $3)}')
	for dev in $phydevs; do
		find_modules_for_devpath $dev
		lvm_ignore_devices="$lvm_ignore_devices $dev"
	done

	local hw hwhandlers=$(echo "$info" | awk '/hwhandler=1/{sub(/.*hwhandler=1 /, ""); sub(/\]$/, ""); print}')
	for hw in $hwhandlers; do
		find_module "dm-$hw"
	done

	local target targets=$(echo "$info" | awk '/prio=/{print $2}' | sort -u)
	for target in $targets; do
		find_module "dm-$target"
	done

	find_module "dm-mod"
	return 0
}

# generate initrd fragment
# @access	public
initrd_gen_multipath() {
	inst_d /sbin /lib/udev /etc/multipath
	inst_exec /sbin/kpartx /sbin
	inst_exec /sbin/multipath /sbin

	# for udev callouts
	inst_exec /sbin/scsi_id /lib/udev
	inst_exec /sbin/mpath* /sbin
	egrep -v '^([ 	]*$|#)' /etc/multipath.conf > $DESTDIR/etc/multipath.conf

	if [ -f /etc/multipath/bindings ]; then
		egrep -v '^([ 	]*$|#)' /etc/multipath/bindings > $DESTDIR/etc/multipath/bindings
	else
		touch $DESTDIR/etc/multipath/bindings
	fi

	mount_dev
	initrd_gen_devices

	mount_sys
	echo "export MPATH_WWID=$MPATH_WWID" | add_linuxrc
	add_linuxrc <<-'EOF'
		# parse mpath_wwid= from kernel commandline
		for arg in $CMDLINE; do
			if [ "${arg##mpath_wwid=}" != "${arg}" ]; then
				MPATH_WWID=${arg##mpath_wwid=}
				if [ "$MPATH_WWID" = "*" ]; then
					# '*' would mean activate all WWID-s
					MPATH_WWID=
					echo "multipath: Activating all WWID-s"
				else
					echo "multipath: Activating WWID=$WWID"
				fi
			fi
		done

		debugshell
		/sbin/multipath -v 0 $MPATH_WWID

		for a in /dev/mapper/*; do
			[ $a = /dev/mapper/control ] && continue
			/sbin/kpartx -a -p p $a
		done
		debugshell
	EOF
}


# PRIVATE METHODS
# export info from dmsetup
# param can be:
# - MAJOR:MINOR
# - /dev/dm-MINOR
# - /dev/mapper/DM_NAME
dm_export() {
	local arg="$1"

	case "$arg" in
	*:*)
		local maj=${arg%:*} min=${arg#*:}
		dmsetup -j $maj -m $min export
		;;
	/dev/dm-*)
		local min=${arg#*dm-}
		local maj=$(awk '$2 == "device-mapper" {print $1}' /proc/devices)
		dm_export $maj:$min
		;;
	/dev/mapper/*)
		local dm_name=${arg#/dev/mapper/}
		dmsetup export $dm_name
		;;
	*)
		die "dm_export: unexpected $arg"
		;;
	esac
}

